//
// AXI bus splitter
// Input: on slave port
// Output: two master ports
//
// Design by G.J. van Loo, FenLogic Ltd, October-2017.
// Mail: verilog@fenlogic.com
//
// This program is free software. It comes without any guarantees or
// warranty to the extent permitted by applicable law. Although the
// author has attempted to find and correct any bugs in this free software
// program, the author is not responsible for any damage or losses of any
// kind caused by the use or misuse of the program. You can redistribute
// the program and or modify it in any form without obligations, but the
// author would appreciated if these credits stays in.
//
// This modules is mostly combinatorial. 
// The efficiency greatkly depends on the response data
// For details read the manual. 
// 
// In the return direction all data is passed in-order. 
// For an explanation why see the documentation.
//

module axi_split3
#(parameter ADRS_BITS  = 32,
            DATA_BITS  = 32,
            ID_BITS    =  6, // Slave port ID Bits 
            LEN_BITS   =  4, // Len bits 
            MASK       = 32'hC000000, // Adddress routing bits
            VALUE0     = 32'h0000000, // Adddress match port 0 
            VALUE1     = 32'h4000000, // Adddress match port 1
            L2MAXTRANS = 3   // Log 2 max outstanding transactions
 )
(
   input              clk    ,
   input              reset_n,
   
   /****************\
    *              *
    *  SLAVE PORT  *
    *              *
   \****************/

  // write address 
   
   input                    s_awvalid , 
   output reg               s_awready , 
   input    [ADRS_BITS-1:0] s_awaddr  , 
   input     [LEN_BITS-1:0] s_awlen   , 
   input              [2:0] s_awsize  , 
   input      [ID_BITS-1:0] s_awid    , 
   input              [1:0] s_awburst , 
   input              [3:0] s_awcache , 
   input              [1:0] s_awlock  , 
   input              [2:0] s_awprot  , 
   input              [3:0] s_awqos   , 
                           
   // write data           
   input                    s_wvalid  , 
   output reg               s_wready  , 
   input    [DATA_BITS-1:0] s_wdata   , 
   input              [3:0] s_wstrb   , 
   input                    s_wlast   , 
   input      [ID_BITS-1:0] s_wid     , 

   // write response    
   output reg               s_bvalid  , 
   input                    s_bready  , 
   output reg         [1:0] s_bresp   , 
   output reg [ID_BITS-1:0] s_bid     , 


   //
   // read interface
   //
   
   // read address 
   input                      s_arvalid, 
   output reg                 s_arready, 
   input      [ADRS_BITS-1:0] s_araddr , 
   input        [ID_BITS-1:0] s_arid   , 
   input       [LEN_BITS-1:0] s_arlen  , 
   input                [2:0] s_arsize , 
   input                [1:0] s_arburst, 
   input                [3:0] s_arcache, 
   input                [1:0] s_arlock , 
   input                [2:0] s_arprot , 
   input                [3:0] s_arqos  , 

   // read data
   output reg                 s_rvalid , 
   input                      s_rready , 
   output reg [DATA_BITS-1:0] s_rdata  , 
   output reg                 s_rlast  , 
   output reg   [ID_BITS-1:0] s_rid    , 
   output reg           [1:0] s_rresp  , 

 
///////////////////////////////////////////////////////////   
 
   /*******************\
    *                 *
    *  MASTER PORT 0  *
    *                 *
   \*******************/
   
   //
   // write interface
   //
   
   // write address 
   
   output reg                 m0_awvalid , 
   input                      m0_awready , 
   output reg [ADRS_BITS-1:0] m0_awaddr  , 
   output reg  [LEN_BITS-1:0] m0_awlen   , 
   output reg           [2:0] m0_awsize  , 
   output reg   [ID_BITS-1:0] m0_awid    , 
   output reg           [1:0] m0_awburst , 
   output reg           [3:0] m0_awcache , 
   output reg           [1:0] m0_awlock  , 
   output reg           [2:0] m0_awprot  , 
   output reg           [3:0] m0_awqos   , 
                           
   // write data           
   output reg                 m0_wvalid  , 
   input                      m0_wready  , 
   output reg [DATA_BITS-1:0] m0_wdata   , 
   output reg           [3:0] m0_wstrb   , 
   output reg                 m0_wlast   , 
   output reg   [ID_BITS-1:0] m0_wid     , 

   // write response    
   input                      m0_bvalid  , 
   output reg                 m0_bready  , 
   input                [1:0] m0_bresp   , 
   input        [ID_BITS-1:0] m0_bid     , 


   //
   // read interface
   //
   
   // read address 
   output reg                 m0_arvalid, 
   input                      m0_arready, 
   output reg [ADRS_BITS-1:0] m0_araddr , 
   output reg   [ID_BITS-1:0] m0_arid   , 
   output reg  [LEN_BITS-1:0] m0_arlen  , 
   output reg           [2:0] m0_arsize , 
   output reg           [1:0] m0_arburst, 
   output reg           [3:0] m0_arcache, 
   output reg           [1:0] m0_arlock , 
   output reg           [2:0] m0_arprot , 
   output reg           [3:0] m0_arqos  , 

   // read data
   input                      m0_rvalid , 
   output reg                 m0_rready , 
   input      [DATA_BITS-1:0] m0_rdata  , 
   input                      m0_rlast  , 
   input        [ID_BITS-1:0] m0_rid    , 
   input                [1:0] m0_rresp  ,
   
   /*******************\
    *                 *
    *  MASTER PORT 1  *
    *                 *
   \*******************/

   //
   // write interface
   //
   
    
   // write address 
   
   output reg                 m1_awvalid , 
   input                      m1_awready , 
   output reg [ADRS_BITS-1:0] m1_awaddr  , 
   output reg  [LEN_BITS-1:0] m1_awlen   , 
   output reg           [2:0] m1_awsize  , 
   output reg   [ID_BITS-1:0] m1_awid    , 
   output reg           [1:0] m1_awburst , 
   output reg           [3:0] m1_awcache , 
   output reg           [1:0] m1_awlock  , 
   output reg           [2:0] m1_awprot  , 
   output reg           [3:0] m1_awqos   , 
                           
   // write data           
   output reg                 m1_wvalid  , 
   input                      m1_wready  , 
   output reg [DATA_BITS-1:0] m1_wdata   , 
   output reg           [3:0] m1_wstrb   , 
   output reg                 m1_wlast   , 
   output reg   [ID_BITS-1:0] m1_wid     , 

   // write response    
   input                      m1_bvalid  , 
   output reg                 m1_bready  , 
   input                [1:0] m1_bresp   , 
   input        [ID_BITS-1:0] m1_bid     , 


   //
   // read interface
   //
   
   // read address 
   output reg                 m1_arvalid, 
   input                      m1_arready, 
   output reg [ADRS_BITS-1:0] m1_araddr , 
   output reg   [ID_BITS-1:0] m1_arid   , 
   output reg  [LEN_BITS-1:0] m1_arlen  , 
   output reg           [2:0] m1_arsize , 
   output reg           [1:0] m1_arburst, 
   output reg           [3:0] m1_arcache, 
   output reg           [1:0] m1_arlock , 
   output reg           [2:0] m1_arprot , 
   output reg           [3:0] m1_arqos  , 

   // read data
   input                      m1_rvalid , 
   output reg                 m1_rready , 
   input      [DATA_BITS-1:0] m1_rdata  , 
   input                      m1_rlast  , 
   input        [ID_BITS-1:0] m1_rid    , 
   input                [1:0] m1_rresp  ,
  
   /*******************\
    *                 *
    *  MASTER PORT 2  *
    *                 *
   \*******************/

   //
   // write interface
   //
   
    
   // write address 
   
   output reg                 m2_awvalid , 
   input                      m2_awready , 
   output reg [ADRS_BITS-1:0] m2_awaddr  , 
   output reg  [LEN_BITS-1:0] m2_awlen   , 
   output reg           [2:0] m2_awsize  , 
   output reg   [ID_BITS-1:0] m2_awid    , 
   output reg           [1:0] m2_awburst , 
   output reg           [3:0] m2_awcache , 
   output reg           [1:0] m2_awlock  , 
   output reg           [2:0] m2_awprot  , 
   output reg           [3:0] m2_awqos   , 
                           
   // write data           
   output reg                 m2_wvalid  , 
   input                      m2_wready  , 
   output reg [DATA_BITS-1:0] m2_wdata   , 
   output reg           [3:0] m2_wstrb   , 
   output reg                 m2_wlast   , 
   output reg   [ID_BITS-1:0] m2_wid     , 

   // write response    
   input                      m2_bvalid  , 
   output reg                 m2_bready  , 
   input                [1:0] m2_bresp   , 
   input        [ID_BITS-1:0] m2_bid     , 


   //
   // read interface
   //
   
   // read address 
   output reg                 m2_arvalid, 
   input                      m2_arready, 
   output reg [ADRS_BITS-1:0] m2_araddr , 
   output reg   [ID_BITS-1:0] m2_arid   , 
   output reg  [LEN_BITS-1:0] m2_arlen  , 
   output reg           [2:0] m2_arsize , 
   output reg           [1:0] m2_arburst, 
   output reg           [3:0] m2_arcache, 
   output reg           [1:0] m2_arlock , 
   output reg           [2:0] m2_arprot , 
   output reg           [3:0] m2_arqos  , 

   // read data
   input                      m2_rvalid , 
   output reg                 m2_rready , 
   input      [DATA_BITS-1:0] m2_rdata  , 
   input                      m2_rlast  , 
   input        [ID_BITS-1:0] m2_rid    , 
   input                [1:0] m2_rresp  

);

  /****************\
   *              *
   *  WRITE PORT  *
   *              *
  \****************/

// 
// For the routing of the data we must have an address. 
// Thus the data ready comes only if there is a valid address.
// This will lock-up if a device outptus data and then waits
// for a ready to output the address.
// (It is possible to work arond that by storing the data
//  until the address appears but that is very expensive.)
// Also the next address is not accepted until all the 
// data has passed. 

localparam PASS_AD = 2'b00,
           PASS_A  = 2'b01,
           PASS_D  = 2'b10;

reg  [1:0]  aw_current_dest; // Destination from address
reg  [1:0]  aw_locked_dest;  // Destination stored
reg  [1:0]  aw_output_port;  // Which port we service 0 or 1
reg  [1:0]  aw_state;        // Where we are 

// Response FIFO signal
wire        bff_write;
wire  [1:0] bff_wdata;
wire        bff_read;
wire  [1:0] bff_rdata;
wire        bff_full;
wire        bff_empty;

wire adrs_end = s_awvalid & s_awready;
wire data_end = s_wvalid  & s_wready & s_wlast;

   // decode write address to output port 
   always @(s_awaddr)
   begin
      if ((s_awaddr & MASK)==VALUE0)
         aw_current_dest = 2'b00;
      else if ((s_awaddr & MASK)==VALUE1)
         aw_current_dest = 2'b01;
      else
         aw_current_dest = 2'b10;
   end
   
   // Switch to locked destination
   // when passing data only  
   always @( * )
   begin
      if (aw_state==PASS_D)
         aw_output_port=aw_locked_dest;
      else
         aw_output_port=aw_current_dest;
   end

   //
   // Write address and data FSM
   // Keeps track of what we have done thusfar   
   // 
   always @(posedge clk or negedge reset_n)
   begin
      if (!reset_n)
      begin
         aw_locked_dest <= 2'b00;
         aw_state       <= PASS_AD;
      end
      else
      begin
 
         case (aw_state)         
         PASS_AD: 
            // address and data are passed on
            // Next state depends on who finishes first
            begin
               case ( {adrs_end,data_end} )
               2'b00 : // Busy passing but no address/data cycles ending
                       // Stay in same state
                     aw_state <= PASS_AD;
               2'b01 : // Done with data. wait for address 
                     aw_state <= PASS_A;
               2'b10 : // Done with address. wait for data 
                  // As address may disappear keep the selected port
                  begin
                     aw_state <= PASS_D;
                     aw_locked_dest <= aw_current_dest;
                  end
               2'b11 : // Both done. 
                     aw_state <= PASS_AD;
               endcase
            end // case PASS_AD
               
         PASS_A: 
            if (adrs_end)
               aw_state <= PASS_AD;
            
         PASS_D:
            if (data_end)
               aw_state <= PASS_AD;
               
         endcase
      end // clocked 
   end // always 


   // 
   // Steer output based on aw_output_port
   // readies are routed back as well but only
   // if the bus (address/data) is active
   // valid and ready are blocked if the response FF is full 
   // 
   // In some cases we COULD reduce the ID bits here 
   // but it is not possible to generalise the conditions
   always @ ( * )
   begin
      
      //
      // Write address busses
      // 
      m0_awvalid = s_awvalid & (aw_output_port==2'b00) & !bff_full;
      m0_awaddr  = s_awaddr ;
      m0_awlen   = s_awlen  ;
      m0_awsize  = s_awsize ;
      m0_awid    = s_awid   ; 
      m0_awburst = s_awburst;
      m0_awcache = s_awcache;
      m0_awlock  = s_awlock ;
      m0_awprot  = s_awprot ;
      m0_awqos   = s_awqos  ;

      m1_awvalid = s_awvalid & (aw_output_port==2'b01) & !bff_full;
      m1_awaddr  = s_awaddr ;
      m1_awlen   = s_awlen  ;
      m1_awsize  = s_awsize ;
      m1_awid    = s_awid   ; 
      m1_awburst = s_awburst;
      m1_awcache = s_awcache;
      m1_awlock  = s_awlock ;
      m1_awprot  = s_awprot ;
      m1_awqos   = s_awqos  ;

      m2_awvalid = s_awvalid & (aw_output_port==2'b10) & !bff_full;
      m2_awaddr  = s_awaddr ;
      m2_awlen   = s_awlen  ;
      m2_awsize  = s_awsize ;
      m2_awid    = s_awid   ; 
      m2_awburst = s_awburst;
      m2_awcache = s_awcache;
      m2_awlock  = s_awlock ;
      m2_awprot  = s_awprot ;
      m2_awqos   = s_awqos  ;

      // No ready for address if we are passing data only 
      // Also not ready if the b-response FIFO is full
      if (aw_state==PASS_D || bff_full)
         s_awready  = 1'b0;
      else
         case (aw_output_port)
         2'b00  : s_awready = m0_awready;
         2'b01  : s_awready = m1_awready;
         default: s_awready = m2_awready;
         endcase

      //
      // Write data busses
      //
      m0_wvalid = s_wvalid & (aw_output_port==2'b00);
      m0_wdata  = s_wdata ;
      m0_wstrb  = s_wstrb ;
      m0_wlast  = s_wlast ;
      m0_wid    = s_wid   ;

      m1_wvalid = s_wvalid & (aw_output_port==2'b01);
      m1_wdata  = s_wdata ;
      m1_wstrb  = s_wstrb ;
      m1_wlast  = s_wlast ;
      m1_wid    = s_wid   ;

      m2_wvalid = s_wvalid & (aw_output_port==2'b10);
      m2_wdata  = s_wdata ;
      m2_wstrb  = s_wstrb ;
      m2_wlast  = s_wlast ;
      m2_wid    = s_wid   ;
      
      // No ready for data if we are passing address only 
      if (aw_state==PASS_A)
         s_wready  = 1'b0;
      else
         case (aw_output_port)
         2'b00  : s_wready = m0_wready;
         2'b01  : s_wready = m1_wready;
         default: s_wready = m2_wready;
         endcase
      
   end // always 

   //
   // Write response handling
   //
   // Store in a FIFO each time data leaves on a port
   // That port is then open for the write response

wire [1:0] bresp_port;

// Using my standard synchronous FIFO
// It is only two bits wide but we need the exact functionality.  
   

   // write to FF each time a write address leaves the splitter
   assign bff_write  = s_awvalid & s_awready;
   // Store the port where it is leaving from 
   assign bff_wdata = aw_output_port;
   
   // read from FF each time a reply leaves the splitter
   assign bff_read  = s_bvalid & s_bready;
   // This is the port we service
   assign bresp_port = bff_rdata;
   // 
   

 sync_fifo
#(.WIDTH(2),            // width in bits
  .L2DEPTH(L2MAXTRANS), // Log 2 Depth, 4=16 deep
  .REGFLAGS(0)          // Full, empty are not registered
)
b_fifo
(
   .clk    (clk      ),   // system clock                 
   .reset_n(reset_n  ), // A-synchronous low reset/clear
   .enable (1'b1     ), // clock gating                 
   .clear  (1'b0     ), // Synchronous clear            
                                            
   .write  (bff_write), // write FIFO                   
   .wdata  (bff_wdata), // write data                   
   .read   (bff_read ), // read FIFO                    
   .rdata  (bff_rdata), // read data                    
         
   .empty  (bff_empty), // FIFO is empty                
   .full   (bff_full ), // FIFO is full                 
   .level  ()            // Fill level                   
);
   
   
   // Select the response input 
   // valid & ready are blocked when 
   // the response FF is empty 
   always @( * )
   begin
      case (bresp_port)
      2'b00 : 
         begin
            s_bvalid = m0_bvalid & !bff_empty;
            s_bresp  = m0_bresp;
            s_bid    = m0_bid;
            
            m0_bready= s_bready  & !bff_empty;
            m1_bready= 1'b0;
            m2_bready= 1'b0;
         end
      2'b01 : 
         begin
            s_bvalid = m1_bvalid & !bff_empty;
            s_bresp  = m1_bresp;
            s_bid    = m1_bid;
            
            m0_bready= 1'b0;
            m1_bready= s_bready  & !bff_empty;
            m2_bready= 1'b0;
         end
      default: 
         begin
            s_bvalid = m2_bvalid & !bff_empty;
            s_bresp  = m2_bresp;
            s_bid    = m2_bid;
            
            m0_bready= 1'b0;
            m1_bready= 1'b0;
            m2_bready= s_bready  & !bff_empty;
         end
      endcase
   end
      
 
  /****************\
   *              *
   *  READ PORTS  *
   *              *
  \****************/

//
// The read port split is much easier
// as we only have to route the address
// 

reg [1:0] ar_output_port;  // Which port we service 0, 1 or 2

wire       rff_write;
wire [1:0] rff_wdata;
wire       rff_read;
wire [1:0] rff_rdata;
wire       rff_full;
wire       rff_empty;

   // decode read address to output port 
   always @(s_araddr)
   begin
      if ((s_araddr & MASK)==VALUE0)
         ar_output_port = 2'b00;
      else if ((s_araddr & MASK)==VALUE1)
         ar_output_port = 2'b01;
      else
         ar_output_port = 2'b10;
   end

   //
   // Read address port mux 
   //
   // valid and ready are blocked if read-return FF is full 
   always @ ( * )
   begin
      // Read address bus 0
      m0_arvalid = s_arvalid & (ar_output_port==2'b00) & !rff_full;
      m0_araddr  = s_araddr ;
      m0_arlen   = s_arlen  ;
      m0_arsize  = s_arsize ;
      m0_arid    = s_arid   ; 
      m0_arburst = s_arburst;
      m0_arcache = s_arcache;
      m0_arlock  = s_arlock ;
      m0_arprot  = s_arprot ;
      m0_arqos   = s_arqos  ;

      // Read address bus 1
      m1_arvalid = s_arvalid & (ar_output_port==2'b01) & !rff_full;
      m1_araddr  = s_araddr ;
      m1_arlen   = s_arlen  ;
      m1_arsize  = s_arsize ;
      m1_arid    = s_arid   ; 
      m1_arburst = s_arburst;
      m1_arcache = s_arcache;
      m1_arlock  = s_arlock ;
      m1_arprot  = s_arprot ;
      m1_arqos   = s_arqos  ;

      // Read address bus 2
      m2_arvalid = s_arvalid & (ar_output_port==2'b10) & !rff_full;
      m2_araddr  = s_araddr ;
      m2_arlen   = s_arlen  ;
      m2_arsize  = s_arsize ;
      m2_arid    = s_arid   ; 
      m2_arburst = s_arburst;
      m2_arcache = s_arcache;
      m2_arlock  = s_arlock ;
      m2_arprot  = s_arprot ;
      m2_arqos   = s_arqos  ;
                
      // And the ready:
      if (rff_full)
         s_arready = 1'b0;
      else
         case (ar_output_port)
         2'b00  : s_arready = m0_arready;
         2'b01  : s_arready = m1_arready;
         default: s_arready = m2_arready;
         endcase
      
   end // always 


   //
   // read response handling
   //
   // Store in a FIFO each time data leaves on a port
   // That port is then open for the read reply


wire [1:0] rreturn_port;

// Using my standard synchronous FIFO
// It is only two bits wide but we need the exact functionality.  
   

   // write to FF each time a read address leaves the splitter
   assign rff_write  = s_arvalid & s_arready;
   // Store the port where it is leaving from 
   assign rff_wdata = ar_output_port;
   
   // read from FF each time a reply leaves the splitter
   assign rff_read  = s_rvalid & s_rready & s_rlast;
   // This is the port we service
   assign rreturn_port = rff_rdata;
   // 
   

 sync_fifo
#(.WIDTH(2),            // width in bits
  .L2DEPTH(L2MAXTRANS), // Log 2 Depth, 4=16 deep
  .REGFLAGS(0)          // Full, empty are not registered
)
r_fifo
(
   .clk    (clk      ), // system clock                 
   .reset_n(reset_n  ), // A-synchronous low reset/clear
   .enable (1'b1     ), // clock gating                 
   .clear  (1'b0     ), // Synchronous clear            
                                            
   .write  (rff_write), // write FIFO                   
   .wdata  (rff_wdata), // write data                   
   .read   (rff_read ), // read FIFO                    
   .rdata  (rff_rdata), // read data                    
         
   .empty  (rff_empty), // FIFO is empty                
   .full   (rff_full ), // FIFO is full                 
   .level  ()            // Fill level                   
);
   
      
   //
   // Select the response input 
   // valid & ready are blocked when the response FF is empty 
   always @( * )
   begin
      case (rreturn_port)
      2'b00 : 
         begin
            s_rvalid = m0_rvalid  & !rff_empty; 
            s_rdata  = m0_rdata ; 
            s_rlast  = m0_rlast ; 
            s_rid    = m0_rid   ; 
            s_rresp  = m0_rresp ; 
   
            m0_rready = s_rready  & !rff_empty;
            m1_rready = 1'b0;
            m2_rready = 1'b0;
         end
      2'b01 : 
         begin
            s_rvalid = m1_rvalid  & !rff_empty; 
            s_rdata  = m1_rdata ; 
            s_rlast  = m1_rlast ; 
            s_rid    = m1_rid   ; 
            s_rresp  = m1_rresp ; 
   
            m0_rready = 1'b0;
            m1_rready = s_rready  & !rff_empty;
            m2_rready = 1'b0;
         end
      default: 
         begin
            s_rvalid = m2_rvalid  & !rff_empty; 
            s_rdata  = m2_rdata ; 
            s_rlast  = m2_rlast ; 
            s_rid    = m2_rid   ; 
            s_rresp  = m2_rresp ; 
   
            m0_rready = 1'b0;
            m1_rready = 1'b0;
            m2_rready = s_rready  & !rff_empty;
         end
      endcase
   end
   
 
endmodule    
